"""
A set of simple tools to build a configuration system with

The first set are 'sources':
  - OSEnvConfigSource: which parses config values from os.environ into hierarchical dictionaries
  - YamlConfigFile: which parses a yaml file into hierarchical dictionaires
The second tool is:
  - MergedConfiguration: which merges multiple configuration sources into a coherent whole

Yes, much more could be done:
    - required keys
    - value validation
    - default parsers
    - default default values (that .get() without a default would default to)
    - more file types (json, ini, env, ... xml?)


Much inspration - and the .get() signature - is taken from everett, but the implementation is all mine.
"""

from .sources import *
from .configs import *




"""
Abstract source definitions, useful for typing and subclassing
"""

from typing import Any, Dict
from abc import ABC, abstractmethod

# CfgDict = Dict[str, Union[str, 'CfgDict']]  # the correct, recursive type... unsupported by mypy
CfgDict = Dict[str, Any]


class ConfigSource(ABC):

    @abstractmethod
    def as_dict(self) -> CfgDict:
        """
        return a dict containing configuration information from this source.

        NOTE: Errors _MUST_ _NOT_ be raised; in case of error, return as much configuration information as possible.
        An empty dict is acceptable (and expected, in cases where, for instance, the supplied configuration file
        doesn't exist)
        """
        raise NotImplementedError



"""
Fake sources for throwing errors
"""

from .abstract import ConfigSource


def ErrorMessagingSource(msg):
    """
    This is used to allow sources with external dependencies to have a value that
    will throw an error if and only if they're actually used
    """

    class ErrMsgSrc(ConfigSource):
        def __init__(self, *a, **kw):
            raise ValueError(msg)

        def as_dict(self):
            raise ValueError(msg)

    return ErrMsgSrc



"""
Configurations are the way that multiple sources are glommed together.

MergedConfiguration merges multiple sources linearly, with later sources taking precedence over earlier ones.

"""

import logging
from copy import deepcopy
from types import SimpleNamespace
from configparser import ConfigParser
from typing import Dict, List, Optional

from .sources import ConfigSource, CfgDict

logger = logging.getLogger(__name__)


def _recursive_dict_update(main: Dict, update: Dict):
    """
    like dict.update(), modifies the 'main' by applying 'update'
    UNLIKE dict.update(), normalizes all keys to lowercase
    :meta private:
    """
    for k, v in update.items():
        lk = k.lower()
        if isinstance(v, dict):
            main[lk] = main.get(lk, dict())
            logger.debug("dict %r %s recursing into dict %r to write %r", id(main), lk, id(main[lk]), v)
            _recursive_dict_update(main[lk], v)
        else:
            logger.debug("dict %r : %s -> %r", id(main), lk, v)
            main[lk] = v


DEFAULT_SENTINEL = object()


class MergedConfiguration:
    """
    Merges configuration sources

    :param sources: a list of :ref:ConfigSource objects to merge, with later ones overriding earlier ones
    """

    def __init__(self, sources: Optional[List[ConfigSource]] = None):
        self.sources: List[ConfigSource] = [] if sources is None else sources
        self._cfg: Optional[CfgDict] = None

    def __repr__(self):
        return 'MergedConfiguration([' + ', '.join(repr(s) for s in self.sources) + ')'

    def add_source(self, source: ConfigSource):
        self.sources.append(source)
        self._cfg = None

    def load(self):
        """
        Cause the config to be loaded; note that this need not be called directly, as it is lazily loaded.
        Subsequent calls _will_ re-load from the config sources.
        """
        cfg: CfgDict = dict()
        for source in self.sources:
            updates = source.as_dict()
            _recursive_dict_update(cfg, updates)
        self._cfg = cfg

    def as_dict(self, namespace: Optional[List[str]] = None) -> CfgDict:
        """
        return the configuration, or a namespace within it, as a single dictionary

        :param namespace: the namespace to return as a dict.  If unspecified, return the entire config.
        """
        if self._cfg is None:
            self.load()
        ns = [] if namespace is None else namespace
        assert self._cfg is not None
        in_ns = deepcopy(self._cfg)
        for subns in ns:
            in_ns = in_ns.setdefault(subns, dict())
        if not isinstance(in_ns, dict):
            raise ValueError(f"namespace {'.'.join(ns)} points to a key, not a dict")
        return in_ns

    def get(self, key: str, namespace: Optional[List[str]]=None, default=DEFAULT_SENTINEL, parser=str, raise_error=True, doc=None):
        """
        get a single config key

        :param namespace: the 'path' in the config to the namespace to look the key up in
        :param default: is the value to use if the key is not found
        :param parser: is how to cast the result
        :param raise_error: is whether to raise an error if the key is not found (note: supplying a default negates
            raise_error, as a (presumably) valid value is always available)
        :param doc: is extra information to supply with the error message
        """
        k = key.lower()
        ns = namespace or []
        in_ns = self.as_dict()
        for subns in ns:
            in_ns = in_ns[subns]
        if k in in_ns:
            return parser(in_ns.get(k, default))
        if default is not DEFAULT_SENTINEL:
            return default
        if raise_error:
            msg = f"Missing config key {'.'.join(ns + [k])!r}"
            if doc:
                msg += f": {doc}"
            raise KeyError(msg)
        return None

    def as_ns(self, namespace: Optional[List[str]]=None) -> SimpleNamespace:
        """
        return the config, or a namespace within it, as a SimpleNamespace

        :param namespace: the namespace to return as a SimpleNamespace.  If unspecified, return the entire config.
        """
        nsdict = self.as_dict(namespace)
        for k, v in nsdict.items():
            if not isinstance(v, dict):
                continue
            nsdict[k] = self.as_ns((namespace or []) + [k])
        return SimpleNamespace(**nsdict)

    def as_configparser(self, namespace: Optional[List[str]]=None) -> ConfigParser:
        """
        return the config, or a namespace within it, as a ConfigParser

        :param namespace: the namespace to return as a ConfigParser.  If unspecified, return the entire config.
        """
        result = ConfigParser()
        for k, v in self.as_dict(namespace).items():
            if isinstance(v, dict):
                result.add_section(k)
                for kk, vv in v.items():
                    result.set(k, kk, vv)
            else:
                result.set(result.default_section, k, v)
        return result



import os.path
from pathlib import Path
from typing import Union

from configobj import ConfigObj  # type: ignore

from .abstract import ConfigSource, CfgDict


class INIFileSource(ConfigSource):
    """
    An INI-file source of configuration information
    Note that if the top level of the file is not a dict, it will be set as the value of an empty key.
    """

    def __init__(self, filename: Union[Path, str]):
        self.filename = str(filename) if filename is not None else None

    def as_dict(self) -> CfgDict:
        if self.filename is None:
            return dict()
        if not os.path.exists(self.filename):
            return dict()
        if not os.path.isfile(self.filename):
            return dict()
        val = ConfigObj(self.filename)
        if not isinstance(val, dict):
            val = {'': val}
        return val



"""
Meta sources - sources that read/filter/fold/spindle/mutilate other sources
"""
from typing import Iterable

from .abstract import ConfigSource, CfgDict


class SubsetSource(ConfigSource):
    """
    Returns a sub-namespace of another source.  Useful with MergedConfiguration if you want partial overrides for
    precedence reasons. eg. source A's namespace  'A' should take priority over source B's namspace 'A', but source B's
    namespace 'B' should take priority over source A's namespace 'B'.
    """

    def __init__(self, source: ConfigSource, keys: Iterable[str]):
        self.source = source
        self.keys = set(keys)

    def as_dict(self) -> CfgDict:
        full = self.source.as_dict()
        return {k: v for k, v in full.items() if k in self.keys}


class SubNamepaceSource(ConfigSource):

    def __init__(self, source: ConfigSource, ns: Iterable[str]):
        self.source = source
        self.ns = set(ns)

    def as_dict(self) -> CfgDict:
        result = self.source.as_dict()
        for subns in self.ns:
            if subns in result:
                result = result[subns]
        return result




import os

from mincfg import MergedConfiguration
from mincfg import DictSource, OSEnvironSource, SubsetSource, YamlFileSource, INIFileSource, DotEnvFileSource


D1 = DictSource({
        'a': '11', 
        'B': '12',
        'C': { 
            'Ca': '131', 
            'CB': '132'
        }
    })

D2 = DictSource({
        'A': '21',
        'c': {
            'ca': '231',
            'cc': '233'
        },
        'd': '24',
    })



def test_merged_case_insensitive():

    config = MergedConfiguration([D1, D2])

    assert config.get('a') == '21'
    assert config.get('b') == '12'
    assert config.get('d') == '24'

    for k in ('a', 'b', 'd'):
        assert config.get(k.lower()) == config.get(k.upper())


def test_as_dict():

    config = MergedConfiguration([D1])

    c = config.as_dict()
    assert c['c']['ca'] == '131'
    assert c['c']['cb'] == '132'

    c = config.as_dict(['c'])
    assert c['ca'] == '131'
    assert c['cb'] == '132'


def test_as_ns():

    config = MergedConfiguration([D1])

    c = config.as_ns()
    assert c.a == '11'
    assert c.b == '12'
    assert c.c.ca == '131'
    assert c.c.cb == '132'

    c = config.as_ns(['c'])
    assert c.ca == '131'
    assert c.cb == '132'


def test_as_configparser():

    config = MergedConfiguration([D1])

    c = config.as_configparser()
    assert c['c']['ca'] == '131'
    assert c['c']['cb'] == '132'
    assert 'c' in c.sections()


def test_osenviron_source():

    cur_keys = set(os.environ.keys())
    os.environ.update({
        'MINCFGTST_A': 'a',
        'MINCFGTST_B': 'b',
        'MINCFGTST_C_A': 'ca',
        'MINCFGTST_C_B': 'cb',
    })

    shell = os.environ.get('SHELL', 'no-shell-defined')
    user = os.environ.get('USER', 'unknown user')

    config = MergedConfiguration([OSEnvironSource('MINCFGTST')])
    assert config.get('a') == 'a'
    assert config.get('b') == 'b'
    assert config.get('a', namespace=['c']) == 'ca'
    assert config.get('b', namespace=['c']) == 'cb'


def test_subset_source():

    a = {'a': '1',
         'b': '1',
         'c': '1'
        }
    b = {'a': '2',
         'b': '2',
         'c': '2'
        }
    a_src = DictSource(a)
    b_src = DictSource(b)
    config = MergedConfiguration([a_src, b_src, SubsetSource(a_src, set('a'))])

    assert config.get('a') == '1'
    assert config.get('b') == '2'
    assert config.get('c') == '2'


def test_yaml_file_source(tmp_path):

    nonexistantsrc = YamlFileSource(None)

    # make up a tempfile name
    cfgfile = tmp_path / "config.yaml"

    # write test config to the temp file
    cfgfile.write_text("a: 1\nb: 2\nc:\n  ca: 3\n  cb: 4\n\n")

    cfgfilesrc = YamlFileSource(str(cfgfile))

    # point the config at it
    config = MergedConfiguration([nonexistantsrc, cfgfilesrc])

    assert config.get('a') == '1'
    assert config.get('b') == '2'
    assert config.get('ca', namespace=['c']) == '3'
    assert config.get('cb', namespace=['c']) == '4'


def test_ini_file_source(tmp_path):

    nonexistantsrc = INIFileSource(None)

    # make up a tempfile name
    cfgfile = tmp_path / "config.ini"

    # write test config to the temp file
    cfgfile.write_text("a = 1\nb=2\n[c]\nca = 3\ncb = 4\n\n")

    cfgfilesrc = INIFileSource(str(cfgfile))

    # point the config at it
    config = MergedConfiguration([nonexistantsrc, cfgfilesrc])

    assert config.get('a') == '1'
    assert config.get('b') == '2'
    assert config.get('ca', namespace=['c']) == '3'
    assert config.get('cb', namespace=['c']) == '4'


def test_dotenv_file_source(tmp_path):

    nonexistantsrc = DotEnvFileSource(None)

    # make up a tempfile name
    cfgfile = tmp_path / "config.env"

    # write test config to the temp file
    cfgfile.write_text("a=1\nb=2\nc_a=3\nc_b=4\n\n")

    cfgfilesrc = DotEnvFileSource(str(cfgfile))

    # point the config at it
    config = MergedConfiguration([nonexistantsrc, cfgfilesrc])

    assert config.get('a') == '1'
    assert config.get('b') == '2'
    assert config.get('a', namespace=['c']) == '3'
    assert config.get('b', namespace=['c']) == '4'






'''
Sources based on python's built-in libraries.
'''

import os
import logging

from .abstract import ConfigSource, CfgDict

logger = logging.getLogger(__name__)


class DictSource(ConfigSource):
    """
    Uses the supplied dict as a config source.  Useful for defaults.
    """

    def __init__(self, cfg_dict: CfgDict):
        self.cfg = cfg_dict

    def __repr__(self):
        return f'DictSource({self.cfg!r})'

    def as_dict(self) -> CfgDict:
        return self.cfg


class KeySplittingDictSource(ConfigSource):
    """
    Uses a dict whose keys are split into a hierarcy as a config source.
    eg. a = 1, b.c = 2, b.d = 3 => { a:1, b: {c:2, d:3} }
    """

    def __init__(self, cfg_dict_lambda, sep: str, prefix: str):
        self.sep = sep
        self.cfg_dict_lambda = cfg_dict_lambda
        self.prefix: str = prefix.strip(sep).upper()

    def as_dict(self) -> CfgDict:
        result: CfgDict = dict()
        input_dict = self.cfg_dict_lambda()
        logger.debug("KeySplittingDictSource transforming dict %r", input_dict)
        for full_key in input_dict:
            if not full_key.startswith(self.prefix):
                continue
            key_path = full_key.split('_')
            if self.prefix:
                key_path = key_path[1:]
            if not key_path:
                continue
            *ns_list, key = key_path
            logger.debug("KeySplittingDictSource: %s -> %r,%s ", full_key, ns_list, key)
            namespace = result
            for subns in ns_list:
                namespace = namespace.setdefault(subns, dict())
            namespace[key] = input_dict[full_key]
        return result


class OSEnvironSource(KeySplittingDictSource):
    """
    Uses os.environ as a config source, by parsing PREFIXd keys into hierarchical dictionaries, splitting on _
    """

    def __init__(self, prefix: str):
        super().__init__(lambda: os.environ, '_', prefix)



from pathlib import Path
from typing import Union

from .abstract import ConfigSource, CfgDict

import yaml


class YamlFileSource(ConfigSource):
    """
    A YAML file source of configuration information.
    Note that if the top level of the yaml file is not a dict, it will be set as the value of an empty key.
    """
    def __init__(self, filename: Union[Path, str]):
        self.path = Path(filename) if filename else None

    def as_dict(self) -> CfgDict:
        if self.path is None:
            return dict()
        if not self.path.exists():
            return dict()
        if not self.path.is_file():
            return dict()
        with self.path.open() as f:
            val = yaml.safe_load(f)
        if not isinstance(val, dict):
            val = { '': val }
        return val



from pathlib import Path
from typing import Union

from dotenv import dotenv_values  # type: ignore

from .builtin import KeySplittingDictSource


class DotEnvFileSource(KeySplittingDictSource):
    """
    A .env-file source of configuration information
    """
    def __init__(self, filename: Union[Path, str], prefix=None):
        key_prefix = '' if prefix is None else prefix        
        super().__init__(lambda: dotenv_values(str(filename)), '_', key_prefix)



from .abstract import CfgDict, ConfigSource
from .builtin import DictSource, OSEnvironSource
from .errsrc import ErrorMessagingSource as ErrSrc
from .meta import SubsetSource

__all__ = [ 
    'CfgDict',
    'ConfigSource',
    'DictSource',
    'OSEnvironSource',
    'YamlFileSource',
    'SubsetSource',
    'INIFileSource',
    'DotEnvFileSource',
]

try:
    from .yamlfilesrc import YamlFileSource
except ImportError:
    YamlFileSource = ErrSrc("Using YamlFileSource requires mincfg[yaml] or mincfg[all] be installed") # type: ignore

try:
    from .inifilesrc import INIFileSource
except ImportError:
    INIFileSource = ErrSrc("Using INIFileSource requries mincfg[ini] or mincfg[all] be installed")  # type: ignore

try:
    from .dotenv import DotEnvFileSource
except ImportError:
    DotEnvFileSource = ErrSrc("Using DotEnvFileSource requires mincfg[env] or mincfg[all] be installed")  # type: ignore
